/* Implementation of MainWindow.
 *
 * Copyright (c) 2003 Ole Laursen.
 *
 * This program is free software; you can redistribute it and/or 
 * modify it under the terms of the GNU General Public License as 
 * published by the Free Software Foundation; either version 2 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307
 * USA.
 */

#include <config.h>

#include <cassert>

#include <gtkmm/dialog.h>
#include <gtkmm/main.h>
#include <gdk/gdkkeysyms.h>

#include "main-window.hpp"

#include "preferences-window.hpp"
#include "new-game-window.hpp"
#include "intro.hpp"
#include "game.hpp"
#include "canvas.hpp"
#include "canvas-message.hpp"
#include "center-menu.hpp"
#include "graphic.hpp"
#include "pixbuf-drawing.hpp"
#include "i18n.hpp"
#include "gui-helpers.hpp"
#include "ucompose.hpp"


MainWindow &MainWindow::instance()
{
  static MainWindow w;
  return w;
}

MainWindow::MainWindow()
  : gconf_dir("/apps/monster-masher")
{
  gconf_client = Gnome::Conf::Client::get_default_client();
  gconf_client->add_dir(gconf_dir);

  ui = get_glade_xml("main_window");
  ui->get_widget("main_window", window);

  ui->get_widget("resurrections_label", resurrections_label);
  ui->get_widget("player1_status_label", player1_status_label);
  ui->get_widget("player2_status_label", player2_status_label);
  ui->get_widget("level_label", level_label);

  resurrections_label->set_text("");
  player1_status_label->set_text("");
  player2_status_label->set_text("");
  level_label->set_text("");
  
  
  // setup canvas
  Gnome::Canvas::Canvas *canvas_ptr;
  ui->get_widget("canvas", canvas_ptr);
  canvas_ptr->property_aa() = true;
  canvas.reset(new Canvas(*canvas_ptr));

  set_canvas_size(make_vector(600, 400), make_vector(0, 0));
  generate_main_menu();

  
  // connect various
  window->signal_key_press_event()
    .connect_notify(SigC::slot(*this, &MainWindow::on_key_press_event));
  window->signal_key_release_event()
    .connect_notify(SigC::slot(*this, &MainWindow::on_key_release_event));

  canvas_ptr->add_events(Gdk::BUTTON_PRESS_MASK);
  canvas_ptr->signal_button_press_event()
    .connect_notify(SigC::slot(*this, &MainWindow::on_button_press_event));

  // connect window closement
  window->signal_delete_event()
    .connect(SigC::slot(*this, &MainWindow::on_closed));

  window->show_all();
}

MainWindow::~MainWindow()
{
  // make sure menu doesn't try to hide message when we are dead
  intro.reset();
  center_menu.reset();
}
  
Gtk::Window &MainWindow::get_window()
{
  return *window;
}

void MainWindow::set_canvas_size(Vector<int> size, Vector<int> offset)
{
  canvas->set_size(size, offset);
  // the window will actually be larger than this because of menus etc.
  window->resize(size.x, size.y);
  
  generate_background(size);
}

void MainWindow::set_resurrections(int resurrections)
{
  resurrections_label->set_text(String::ucompose("Resurrections: %1",
						 resurrections));
}

void MainWindow::set_player_info(int player_no, int score)
{
  Glib::ustring s = String::ucompose("Score: %1", score);
  
  if (player_no == 1)
    player1_status_label->set_text(s);
  else
    player2_status_label->set_text(s);
}

void MainWindow::set_level_info(int level)
{
  level_label->set_text(String::ucompose("Level: %1", level));
}

void MainWindow::display_message(const Glib::ustring &message, bool centered)
{
  if (!canvas_message.get())
    canvas_message.reset(new CanvasMessage);
  
  canvas_message->set_message(message, centered);
  canvas_message->show();
}

void MainWindow::remove_message()
{
  if (canvas_message.get())
    canvas_message->hide();
}

void MainWindow::start_new_game(int start_level, Vector<int> size)
{
  // make sure no old info is hanging around
  resurrections_label->set_text("");
  player1_status_label->set_text("");
  player2_status_label->set_text("");
  level_label->set_text("");
  
  center_menu.reset();
  game.reset(); 		// make sure old is gone before creating new
  game.reset(new Game(start_level, size));
}

void MainWindow::end_game()
{
  remove_message();
  game.reset();

  generate_main_menu();
}

void MainWindow::intro_finished()
{
  intro.reset();
  generate_main_menu();
}

void MainWindow::generate_background(Vector<int> size)
{
  Glib::RefPtr<Gdk::Pixbuf> p
    = Gdk::Pixbuf::create(Gdk::COLORSPACE_RGB, false, 8, size.x, size.y);

  for (PixelIterator i = begin(p), e = end(p); i != e; ++i) {
    i->red() = 200 + std::rand() % 30 - 15;
    i->blue() = 200 + std::rand() % 30 - 15;
    i->green() = 200 + std::rand() % 30 - 15;
  }
  
  background.reset(new Graphic(canvas->background_layer));
  background->set_pixbuf(p);

  background->place_at(canvas->get_offset());
}

void MainWindow::generate_main_menu()
{
  CenterMenu::action_sequence actions;
  CenterMenu::Action a;

  a.label = _("_New game");
  a.callback = SigC::slot(*this, &MainWindow::on_new_game_activated);
  actions.push_back(a);
  a.label = _("_Introduction");
  a.callback = SigC::slot(*this, &MainWindow::on_introduction_activated);
  actions.push_back(a);
  a.label = _("_Preferences");
  a.callback = SigC::slot(*this, &MainWindow::on_preferences_activated);
  actions.push_back(a);
  a.label = _("_About");
  a.callback = SigC::slot(*this, &MainWindow::on_about_activated);
  actions.push_back(a);
  a.label = _("_Quit");
  a.callback = SigC::slot(*this, &MainWindow::on_quit_activated);
  actions.push_back(a);
  
  center_menu.reset(new CenterMenu(actions));
}

void MainWindow::generate_game_menu()
{
  CenterMenu::action_sequence actions;
  CenterMenu::Action a;

  a.label = _("_Resume game");
  a.callback = SigC::slot(*this, &MainWindow::on_resume_game_activated);
  actions.push_back(a);
  a.label = _("Re_start level");
  a.callback = SigC::slot(*this, &MainWindow::on_restart_level_activated);
  actions.push_back(a);
  a.label = _("_End game");
  a.callback = SigC::slot(*this, &MainWindow::on_end_game_activated);
  actions.push_back(a);
  
  center_menu.reset(new CenterMenu(actions));
}



// call-backs

void MainWindow::on_new_game_activated()
{
  if (new_game_window.get() == 0) {
    new_game_window.reset(new NewGameWindow);
    new_game_window->show();
  }
  else {
    new_game_window->show();
    new_game_window->raise();
  }
}

void MainWindow::on_resume_game_activated()
{
  game->unpause();
  center_menu.reset();
}

void MainWindow::on_restart_level_activated()
{
  center_menu.reset();
  start_new_game(game->get_level(), game->get_size());
}

void MainWindow::on_end_game_activated()
{
  end_game();
}

void MainWindow::on_introduction_activated()
{
  center_menu.reset();
  intro.reset(new Intro);
}

void MainWindow::on_preferences_activated()
{
  if (preferences_window.get() == 0) {
    preferences_window.reset(new PreferencesWindow);
    preferences_window->show();
  }
  else {
    preferences_window->show();
    preferences_window->raise();
  }
}

void MainWindow::on_quit_activated()
{
  window->hide();
}

void MainWindow::on_about_activated()
{
  if (about.get() == 0) {
    std::vector<Glib::ustring> authors;
    authors.push_back("Ole Laursen <olau@hardworking.dk>");
  
    std::vector<Glib::ustring> documenters;
    // add documenters here

    // note to translators: please fill in your names and email addresses
    Glib::ustring translators = _("translator-credits");

    Glib::ustring description =
      _("Fast-paced monster mashing game for GNOME. Splat the monsters "
	"before they roll over you and take over the world of the gnomes!");
  
    about.reset(new Gnome::UI::About(_("Monster Masher"), VERSION,
				     _("Copyright (c) 2003 Ole Laursen"),
				     authors, documenters, description,
				     translators == "translator-credits"
				     ? "" : translators));
    //, Glib::RefPtr< Gdk::Pixbuf > &logo)
    about->set_transient_for(*window);
    about->show();
  }
  else {
    about->show();
    about->raise();
  }
}

void MainWindow::on_key_press_event(GdkEventKey *event)
{
  if (intro.get())
    intro->key_pressed(event->keyval);
  else if (game.get()) {
    if (event->keyval == GDK_Escape)
      if (center_menu.get())
	on_resume_game_activated();
      else {
	game->pause();
	generate_game_menu();
      }
    
    game->key_pressed(event->keyval);
  }
}

void MainWindow::on_key_release_event(GdkEventKey *event)
{
  if (game.get())
    game->key_released(event->keyval);
}

void MainWindow::on_button_press_event(GdkEventButton* event)
{
  if (game.get())
    if (center_menu.get())
      on_resume_game_activated();
    else {
      game->pause();
      generate_game_menu();
    }
}

bool MainWindow::on_closed(GdkEventAny *)
{
  return false;
}
